/*
 * undump (file) <file.dump
 *
 * 92/09/08 StG initial coding
 * 96/01/22 MSH Added support for hdump (hex dumps and LFs)
 *              usage(), Option parsing, etc.
 */

#include <stdio.h>

#include "cnv.h"

#define ERR (-1)
extern int errno;

char          linein[80];
char          *acLine;
unsigned char acOut[16];
unsigned long lSync = 0;

char *usgmsg[] = {
	"Syntax: undump [<opts>] outfile [<opts>] <infile\n",
	"Function: recreate file from hdump output\n",
	"Option\n",
	"   -h   help\n",
	"   -v   version\n",
	NULL
};

version(name)
char *name;
{
	fprintf(stderr,"%s: Version %d.%d %s\n",name,VERSION,REVISION,RELEASE);
}

usage()
{
	int i = 0;

	while (usgmsg[i] != NULL)
		fputs(usgmsg[i++],stderr);

	exit (0);
}

main(argc,argv)
int   argc;
char *argv[];
{
	FILE *ofp;
	int  i;
	int  iTmp;
	char *fname = NULL;
	char *opt;

	if (argc < 2)
		usage();

	for (i = 1; i < argc ; i++)
	{
		if (argv[i][0] == '-')
		{
			opt = &argv[i][1];
			while (*opt)
			{
				switch (toupper(*opt++))
				{
				case 'V':
					version(argv[0]);
				case 'H':
				default :
					usage();
				}
			}
		}
		else
			fname = argv[i];
	}

	if (fname == NULL)
		usage();


	if ((ofp = fopen(fname,"w")) == (FILE *)NULL)
	{
		fprintf(stderr,"undump: error creating file\n");
		exit(errno);
	}

	acLine = linein;
	while (fgets(acLine,80,stdin) != (char *)NULL)
	{
		if (*acLine == 0x0D || *acLine == 0x0A) continue;
		if (!strncmp(acLine,"  Addr",6)) continue;
		if (!strncmp(acLine,"------",6)) continue;
		if (!strncmp(acLine,"  Offs",6)) continue;
		if (!strncmp(acLine," -----",6)) continue;
		if (!strncmp(acLine,"         ***",12))
		{
			iTmp = number(acLine + 12);
			lSync += iTmp * 16;
			while (iTmp--)
				if (fwrite(acOut,16,1,ofp) == ERR)
					exit(errno);
			continue;
		}
		if (*acLine == ' ')
			acLine++;

		if (*acLine < '0' || (*acLine > '9' && *acLine < 'A') || *acLine > 'F')
		{
			fprintf(stderr,"undump: unknown format:\n");
			fprintf(stderr,acLine);
			exit(1);
		}
		iTmp = convert(acLine);
		if (!iTmp)
		{
			fprintf(stderr,"undump: error in line:\n");
			fprintf(stderr,acLine);
			exit(1);
		}
		lSync += iTmp;
		if (fwrite(acOut,iTmp,1,ofp)==ERR)
			exit(errno);

		if (iTmp != 16) break;
		acLine = linein;
	}
}

number(s)
char *s;
{
	int n = 0;

	while (*s == ' ')
		s++;
	while (*s)
	{
		if (*s < '0')
			break;

		if (*s > '9')
			break;

		n *= 10;
		n += *s - '0';
		s++;
	}
	return(n);
}

convert(s)
char *s;
{
	unsigned long lAddr;
	int n;

	n = 4;
	while (n--)
	{
		lAddr <<= 8;
		lAddr |= hex(s);
		s += 2;
	}

	if (lAddr != lSync)
	{
		fprintf(stderr,"undump: address sync error\n");
		fprintf(stderr,"undump: got %08x wanted %08x\n",lAddr,lSync);
		exit(1);
	}

	if (*s++ != ' ')
		return(0);

	if (*s++ != ' ')
		return(0);

	n = 0;
	while (n < 16)
	{
		if (*s == ' ')
			break;
		acOut[n] = hex(s);
		s += 2;
		if (n & 1)
			if (*s++ != ' ')
				return(0);
		n++;
	}
	return(n);
}

hex(s)
char *s;
{
	return(nyb(*s) << 4 | nyb(*(s + 1)));
}

nyb(c)
int c;
{
	if (c < '0' || (c > '9' && c < 'a') || c > 'f')
	{
		fprintf(stderr,"undump: bad hex value\n");
		exit(1);
	}
	if (c > '9')
		return(c - 'a' + 10);
	else
		return(c - '0');
}

